; Nigel Goodwin 1 December 2001
; PWM routines incorporated in my radio test program,
; in keeping with the supplied code I've added a macro for
; my delay routine, plus page select macros.
;
; Cybot test file
; v0.3
;
; David A Oram
; 28 October 2001
;
; PWM Routine based around techbot code by Jerry Merrill
;	with modifications for Cybot H-Bridge design
;	which has OFF/FWD & OFF/REV inputs for each motor
;	whereas techbot has OFF/ON & FWD/REV 

	list	P=16C84
	include <P16C84.inc>
	__config H'3FF9'

;Program Variables stored in SRAM

	cblock	0x0C	;Start of SRAM
	d1, d2, del	;delay variables
	Bit_Cntr        ;bit counter for RS232
	Rcv_Byte	;byte received
	Delay_Count	;delay variable for RS232
	rnd		;counter for random left/right
	wsave, ssave	;Temp register store
	upcnt1, dwncnt1	;4 counters for PWM routine
	upcnt2, dwncnt2
	lftduty, rtduty	;Duty cycle for each motor
	mstatus		;motor activation register
			; Bit 0 = Left Reverse
			; Bit 1 = Left Forward
			; Bit 2 = Right Reverse
			; Bit 3 = Right Forward
	mmask		;mask for motor control
			; Used in PWM routine

	endc
;
;=====================================================	

;Motor Speed Constants

speed_stop	equ	0
speed_slow	equ	1
speed_2		equ	2
speed_3		equ	3
speed_4		equ	4
speed_half	equ	5
speed_6		equ	6
speed_7		equ	7
speed_8		equ	8
speed_9		equ	9
speed_full	equ	D'10'

Ser_In          Equ     04	;bit position for RS232 in


onmsk	equ	B'11111111'	;Motor on Mask
offmsk1	equ	B'11111100'	;Motor1 off mask
offmsk2	equ	B'11110011'	;Motor2 off mask
motor	equ	PORTB	;Motor Port assignment

;	Decimal equvalents of motor assignment bits
lft_rev	equ	1
lft_fwd	equ	2	
rt_rev	equ	4
rt_fwd	equ	8

;Motor Direction Constants

forward		equ	lft_fwd+rt_fwd	;Move Forward
reverse		equ	lft_rev+rt_rev	;Move Reverse
rot_fwd_left	equ	rt_fwd
rot_fwd_right	equ	lft_fwd
rot_rev_left	equ	rt_rev
rot_rev_right	equ	lft_rev
spin_aclkwise	equ	rt_fwd+lft_rev
spin_clkwise	equ	lft_fwd+rt_rev

;=====================================================

;Macros

page1		macro
		bsf	STATUS,RP0
		endm

page0		macro
		bcf	STATUS,RP0
		endm

action	macro	direction,leftspeed,rightspeed

	local	spd=leftspeed+D'16'*rightspeed
	movlw	spd
	call	setspd
	
	movlw	direction
	movwf	mstatus

	endm

; Time delay macro, call wait [number of seconds] - max 25.5 seconds

wait	macro	variable
	movlw	variable*0x0a
	call	Delay

	endm

;+++++
;	BREVC tests the variable in register 'VAR1' and if it is equal to the constant
;	'CONST', branches to label 'LABEL'.

;	Call as:	BREVC	THIS,THAT,THERE

;	Action:		Tests if THIS = THAT and if they are equal, branches to THERE




BREVC	MACRO	VAR1,CONST,LABEL
	MOVLW	CONST
	SUBWF	VAR1,W
	BZ	LABEL		; Branch if VAR1 = CONST
	ENDM


;=====================================================

;Start of Code

	org	0x00	;Reset vector

	nop
	goto	start

;motor control routine
;runs on each timer0 interrupt
;
;needs to be passed duty cycle for each motor
; in registers lftduty and rtduty
;Value of duty cycle is from 0 (stop) to 10 (Full speed)
;Also mstatus to contain details of motor directions
;
;Saves contents of w and status registers
;
	org	0x04	;interrupt vector

isr	movwf	wsave	;save accumulator
	movfw	STATUS
	movwf	ssave
	bcf	STATUS,RP0
			;now do first motor
	movfw	upcnt1
	btfsc	STATUS,Z
	goto	dodwn1
	decf	upcnt1,f
	goto	do2nd
dodwn1	movfw	dwncnt1
	btfsc	STATUS,Z
	goto	doover1
	decf	dwncnt1,f

	bcf	mmask,0	;motor 1 off
	bcf	mmask,1

	goto	do2nd
doover1	movfw	lftduty
	movwf	dwncnt1
	comf	dwncnt1,f
	movlw	11
	addwf	dwncnt1,f
	movfw	lftduty
	movwf	upcnt1
	btfss	STATUS,Z
	goto	mot1on
	goto	do2nd

mot1on	bsf	mmask,0	;turn on motor
	bsf	mmask,1

			;now do second motor
do2nd	movfw	upcnt2
	btfsc	STATUS,Z
	goto	dodown2
	decf	upcnt2,f
	goto	done
dodown2	movfw	dwncnt2
	btfsc	STATUS,Z
	goto	doover2
	decf	dwncnt2,f

	bcf	mmask,2	;turn motor 2 off
	bcf	mmask,3

	goto	done
doover2	movfw	rtduty
	movwf	dwncnt2
	comf	dwncnt2,f
	movlw	11
	addwf	dwncnt2,f
	movfw	rtduty
	movwf	upcnt2
	btfss	STATUS,Z
	goto	mot2on
	goto	done
		
mot2on	bsf	mmask,2	;motor 2 on
	bsf	mmask,3

done	movfw	mmask		;Output to motors
	andwf	mstatus,w
	movwf	motor

	incf	rnd, f		;increment random counter

	bcf	INTCON,T0IF	;clear timer interrupt
	movfw	ssave		;restore w and status
	movwf	STATUS
	movfw	wsave
	retfie			;return from interrupt

;======================================================
; Subroutines
;	setspd	-Set Motor Speeds
;	Delay	-Delay 'W' 10ths of a second


;Set Motor Speeds
;	w register contains left and right speed
;		left speed in low nibble
;		right speed in high nibble
;
; This routine could probably be dispensed with
;	and left/right speeds passed directly

setspd	movwf	lftduty
	movwf	rtduty
	swapf	rtduty,f
	movlw	B'00001111'
	andwf	lftduty,f
	andwf	rtduty,f
	return


; call 100mS delay number of times in W
Delay	movwf	del
DelLoop	call 	Del100
	decfsz	del, f
	goto	DelLoop
	return

; 100mS delay
Del100	movlw	0x8d
	movwf	d1
Delay_00
	movlw	0xeb
	movwf	d2
Delay_01
	decfsz	d2, f
	goto	Delay_01
	decfsz	d1, f
	goto	Delay_00

			;25 cycles
	movlw	0x08
	movwf	d1
Delay_10
	decfsz	d1, f
	goto	Delay_10

			;1 cycles
	nop

			;4 cycles (including call)
	return

STOP		; stop both motors
	BCF	PORTB,	5
	action	forward,speed_stop,speed_stop
	return

FORWARD		; both motors forward
	BCF	PORTB,	5
	action	forward,speed_3,speed_3
	return

BACKWARD	; both motors backward
	BCF	PORTB,	5
	action	reverse,speed_3,speed_3
	return

LEFT		; RH motor forwards
	BCF	PORTB,	5
	action	rot_fwd_left,speed_slow,speed_slow
	return

RIGHT		; LH motor forwards
	BCF	PORTB,	5
	action	rot_fwd_right,speed_slow,speed_slow
	return

SPINL		; RH motor forwards, LH motor backwards
	BCF	PORTB,	5
	action	spin_aclkwise,speed_slow,speed_slow
	return

SPINR		; LH motor forwards, RH motor backwards
	BCF	PORTB,	5
	action	spin_clkwise,speed_slow,speed_slow
	return

FORW_F		; both motors forward
	BCF	PORTB,	5
	action	forward,speed_6,speed_6
	return

BACK_F		; both motors backward
	BCF	PORTB,	5
	action	reverse,speed_6,speed_6
	return

LEFT_F		; RH motor forwards
	BCF	PORTB,	5
	action	rot_fwd_left,speed_slow,speed_2
	return

RIGHT_F		; LH motor forwards
	BCF	PORTB,	5
	action	rot_fwd_right,speed_slow,speed_2
	return

SPINL_F		; RH motor forwards, LH motor backwards
	BCF	PORTB,	5
	action	spin_aclkwise,speed_2,speed_2
	return

SPINR_F		; LH motor forwards, RH motor backwards
	BCF	PORTB,	5
	action	spin_clkwise,speed_2,speed_2
	return

BOTH
	BTFSC	rnd, 0	;check if odd or even
	GOTO	ANTIC
	action	spin_clkwise,speed_slow,speed_slow
wait1	BTFSC   PORTB,	7
	GOTO	wait1	;spin until opposite sensor clear
	return
ANTIC	action	spin_aclkwise,speed_slow,speed_slow
wait2	BTFSC   PORTB,	6
	GOTO	wait2	;spin until opposite sensor clear
	return

DODGEL		; blocked to left
	BTFSC   PORTB,	6	; check if both sensors triggered!
	GOTO	BOTH
	action	rot_fwd_right,speed_2,speed_2
	return

DODGER		; blocked to right
	action	rot_fwd_left,speed_2,speed_2
	return

RADAR		; both motors forward
	BSF	PORTB,	5
	BTFSC   PORTB,	7
	GOTO	DODGEL
	BTFSC   PORTB,	6
	GOTO	DODGER
	action	forward,speed_2,speed_2
	return

; wait for long low before reading RS232
Chk_RS232	MOVLW	20			;check for 20 half bit times
		MOVWF	Bit_Cntr		;use same bit counter as main routine
Chk_Loop	BTFSC   PORTB     , Ser_In
		GOTO	Chk_RS232
		CALL	Start_Delay
		DECFSZ  Bit_Cntr  , f
		GOTO	Chk_Loop

;input byte fron Ser_In at 9600 baud
Rcv_RS232       BTFSS   PORTB     , Ser_In	;wait for start bit
                GOTO    Rcv_RS232		;do half bit time delay
                CALL    Start_Delay
                BTFSS   PORTB     , Ser_In	;confirm still in start bit
                GOTO    Rcv_RS232
                MOVLW   0x08			;set up to read 8 bits
                MOVWF   Bit_Cntr
                CLRF    Rcv_Byte
Next_RcvBit     CALL    Bit_Delay
                BTFSC   PORTB     , Ser_In
                BCF     STATUS    , C
                BTFSS   PORTB     , Ser_In
                BSF     STATUS    , C
                RRF     Rcv_Byte  , f
                DECFSZ  Bit_Cntr  , f
                GOTO    Next_RcvBit
                CALL    Bit_Delay
		CALL    Start_Delay
                BTFSC   PORTB     , Ser_In	;check valid stop bit.
		GOTO    Rcv_RS232
                RETURN

Start_Delay     MOVLW   0x0C
                MOVWF   Delay_Count
Start_Wait      NOP
                DECFSZ  Delay_Count , f
                GOTO    Start_Wait
                RETURN

Bit_Delay       MOVLW   0x18
                MOVWF   Delay_Count
Bit_Wait        NOP
                DECFSZ  Delay_Count , f
                GOTO    Bit_Wait
                RETURN


Chk_Ser
	call	Chk_RS232
	BREVC	Rcv_Byte, 'S', STOP
	BREVC	Rcv_Byte, 'F', FORWARD
	BREVC	Rcv_Byte, 'B', BACKWARD
	BREVC	Rcv_Byte, 'L', LEFT
	BREVC	Rcv_Byte, 'R', RIGHT
	BREVC	Rcv_Byte, 'A', SPINL
	BREVC	Rcv_Byte, 'C', SPINR
	BREVC	Rcv_Byte, 'f', FORW_F
	BREVC	Rcv_Byte, 'b', BACK_F
	BREVC	Rcv_Byte, 'l', LEFT_F
	BREVC	Rcv_Byte, 'r', RIGHT_F
	BREVC	Rcv_Byte, 'a', SPINL_F
	BREVC	Rcv_Byte, 'c', SPINR_F
	BREVC	Rcv_Byte, 'I', RADAR
	BREVC	Rcv_Byte, 0x00, STOP
	return

;======================================================

start	page0			;initialise i/o direction
	clrf	motor
	clrwdt
	page1
	movlw	B'11010000'	;Bits 0-3, and 5 are outputs
	movwf	TRISB
	movlw	1
	movwf	OPTION_REG	;prescaler /4
	movlw	B'11011111'
	andwf	OPTION_REG,f
	movlw	#0A0h
	movwf	INTCON
	page0
	movlw	0		;zero PWM registers
	movwf	upcnt1
	movwf	dwncnt1
	movwf	upcnt2
	movwf	dwncnt2
	BCF	PORTB,	5
	
	action	forward,speed_stop,speed_stop	;initial stop mode
Loop	call Chk_Ser
	goto	Loop	;loop back to start and run for ever.

	end
